# -*- coding: utf-8 -*-
from . import stacker, XLogerBase
from flask import request
from .client import XLogerClient
import time
import uuid
import re
import urllib


class FlaskXLoger(XLogerBase):
    def __init__(self, app=None, config_prefix='XLOGER', **kwargs):
        self.client = XLogerClient
        self.provider_kwargs = kwargs
        self.config_prefix = config_prefix
        stacker['xloger'] = self
        if app is not None:
            self.init_app(app)

    def init_app(self, app, **kwargs):
        host = app.config.get('{0}_HOST'.format(self.config_prefix), "localhost")
        port = app.config.get('{0}_PORT'.format(self.config_prefix), "19527")
        backend = app.config.get('{0}_FILTER_BACKEND'.format(self.config_prefix), "file:///tmp/xloger.filter")
        self.client.config(host, port, filter_backend=backend)
        that = self

        @app.before_request
        def before_request():
            request.xloger_thread = that.thread()
            request.xloger_time_start = time.time()
            request.xloger_thread_data = that.thread_data()
            that.is_watched() and that.trace("threadStart", request.xloger_thread_data)

        @app.teardown_request
        def teardown_request(fn):
            tdata = getattr(request, 'xloger_thread_data', None)
            if not tdata:
                return
            tdata.update(duration=time.time()- getattr(request, "xloger_time_start", time.time()))
            that.is_watched() and that.trace("threadEnd", tdata)

    def log(self, *args):
        if not self.is_watched():
            return
        return self.traceback('log', self.traceback_point(*args))

    def thread_data(self):
        headers = request.headers
        post = ""
        content_type = getattr(request, 'content_type', None)
        if content_type and content_type.lower() == 'application/x-www-form-urlencoded':
            post = urllib.urlencode(request.form)
        else:
            post = request.data

        return dict(
            thread=getattr(request, 'xloger_thread', self.thread()),
            timestamp=time.time(),
            host=headers.get("Host", "localhost"),
            userAgent=headers.get("User-Agent", "none"),
            clientIP=self.clientip(),
            httpMethod=request.method,
            postData=post,
            requestURI=request.full_path,
            cookie=headers.get("Cookie", '')
        )

    @staticmethod
    def thread():
        headers = request.headers
        thread = uuid.uuid1().hex
        super_thread = None
        for hv, qv in (("XLoger-Thread", "xloger_thread"), ('Console-Thread', 'console_thread')):
            super_thread = headers.get(hv, request.values.get(qv, None))
            if super_thread is not None:
                break
        return '_'.join([super_thread, thread]) if super_thread else thread

    @staticmethod
    def clientip():
        headers = request.headers
        for h in ("Client-IP", "X-Real-IP", "Remote-Addr"):
            ip = headers.get(h, None)
            if ip:
                return ip
        xff = headers.getlist("X-Forwarded-For")
        if xff:
            return xff[0]
        return request.remote_addr

    def is_watched(self):
        if hasattr(request, "xloger_watched"):
            return request.xloger_watched

        tdata = getattr(request, 'xloger_thread_data', self.thread_data())
        filter = self.client.filter()
        filters = filter.get("list", [])
        server_mention = filter.get("server_mention", False)

        watched = False
        for f in filters:
            single_watched = True
            fkeys = f.items()
            for k, v in fkeys:
                exp = v.replace(".", '\.').replace("*", ".*").replace("/", "\/");
                rex = re.compile(exp, re.IGNORECASE)
                k = k.lower()
                if k == "serverip":
                    if not server_mention:
                        single_watched = False
                        break

                if k == "clientip":
                    if not rex.match(tdata['clientIP']):
                        single_watched = False
                        break

                if k == "host":
                    if not rex.match(tdata['host']):
                        single_watched = False
                        break

                if k == "useragent":
                    if not rex.match(tdata['userAgent']):
                        single_watched = False
                        break

                if k == "httpmethod":
                    if not rex.match(tdata['httpMethod']):
                        single_watched = False
                        break
                if k == "requesturi":
                    if not rex.match(tdata['requestURI']):
                        single_watched = False
                        break
            if len(fkeys) == 0:
                single_watched = False

            if single_watched:
                watched = True
                break

        request.xloger_watched = watched
        return watched







# -*- coding: utf-8 -*-
from socketpool import ConnectionPool, Connector, TcpConnector
import socket
import select
import errno
import time
from time import sleep
import random
import json
import warnings
import os


class XLogerConnector(TcpConnector):
    def sendall(self, data, flags=0):
        return self._s.send(data, flags)

    pass


class XLogerClient(object):
    """

    """
    host = None
    port = None
    pool = None
    filter_backend = None

    @classmethod
    def config(cls, host="localhost", port=19100, factory=XLogerConnector, filter_backend='file:///tmp/xloger.filter'):
        cls.host = host
        cls.port = port
        cls.pool = ConnectionPool(factory=factory)
        cls.filter_backend = filter_backend

    @classmethod
    def start_filter_worker(cls, host, port, filter_backend='file:///tmp/xloger.filter', retry=0):

        try:
            receiver = socket.create_connection((host, port))
        except socket.error as err:
            warnings.warn("Connect to XLoger Server Failed: [%s] %s" % (err[0], err[1]))
            sleep(3)
            cls.start_filter_worker(host, port, filter_backend, retry+1)
            return

        cls.filter_backend = filter_backend
        # non-blocking
        receiver.setblocking(0)
        data = dict(action="register", data=dict(duplex=True, reciever=True))
        receiver.send(json.dumps(data)+'\n')

        def reconnect():
            receiver.close()
            sleep(3)
            cls.start_filter_worker(host, port, filter_backend)

        while True:
            try:
                line = receiver.makefile().readline()
            except socket.error as e:
                err = e.args[0]
                if err == errno.EAGAIN or err == errno.EWOULDBLOCK:
                    sleep(1)
                    continue
                if err == errno.ECONNRESET:
                    reconnect()
                    break
                continue

            if line:
                cls.handle_package_data(line)
            else:
                try:
                    receiver.send(json.dumps(data)+'\n')
                except socket.error as e:
                    if e.errno == errno.ECONNRESET:
                        reconnect()
                        break

    @classmethod
    def send(cls, data):
        try:
            conn = cls.pool.get(host=cls.host, port=cls.port)
            conn.sendall(data)
        except Exception as e:
            print("XLoger push failed. %s" % e)

    @classmethod
    def push(cls, action='log',  data=''):
        data = dict(action=action, data=data)
        stream = json.dumps(data)+'\n'
        cls.send(stream)

    @classmethod
    def handle_package_data(cls, data):
        try:
            data = json.loads(data)
        except Exception as e:
            warnings.warn("Invalid data recieved.")
            return
        action, data = data.get("action", None), data.get("data", None)
        if not action or not isinstance(data, dict):
            warnings.warn("Invalid data recieved.")
            return
        cls.dispatch(action, data)

    @classmethod
    def dispatch(cls, action, data):
        action = getattr(cls, "dispatch_"+action, None)
        callable(action) and action(data)

    @classmethod
    def dispatch_filter(cls, filter):
        backend = cls.filter_backend
        if backend.startswith("file://"):
            try:
                filter_file_name = backend[7:]
                f = open(filter_file_name, "w+")
                f.write(json.dumps(filter))
                f.close()
            except Exception as e:
                warnings.warn("Failed to Write Xloger backend: %s" % e.message)

    @classmethod
    def filter(cls):
        backend = cls.filter_backend
        if backend.startswith("file://"):
            filter_file_name = backend[7:]
            try:
                f = open(filter_file_name, 'r+')
                data = f.read()
                if data:
                    return json.loads(data)
            except Exception as e:
                warnings.warn("Failed to read Xloger backend: %s" % e.message)
        return dict()







# -*- coding: utf-8 -*-
from functools import partial
from werkzeug.local import LocalStack, LocalProxy
from traceback import extract_tb, extract_stack
from sys import exc_info
import json
import time, datetime

stacker = {"xloger": None}


def _lookup_(name):
    return stacker[name]

xloger = LocalProxy(partial(_lookup_, "xloger"))


def dumps_default(o):
    if hasattr(o, '__json__') and callable(getattr(o, '__json__')):
        return o.__json__()

    if hasattr(o, '__dict__'):
        return o.__dict__

    if isinstance(o, datetime):
        return o.strftime("%Y-%m-%d %H:%M:%S")

    if isinstance(o, time):
        return o.strftime("%H:%M:%S")

    return "%s" % o


class XLogerBase(object):

    def clientip(self):
        raise NotImplementedError()

    def thread(self):
        raise NotImplementedError()

    def thread_data(self):
        raise NotImplementedError()

    def log(self, *args):
        return self.traceback('log', self.traceback_point(*args))

    def warning(self, *args):
        return self.traceback('warning', self.traceback_point(*args))

    def error(self, *args):
        return self.traceback('error', self.traceback_point(*args))

    def show_error(self, errtype, message, file, line):
        data = dict(
            file=file,
            line=line,
            message="%s: %s" % (errtype, message),
            args=[]
        )
        return self.traceback('error', data)

    def traceback(self, ttype, data):
        fire = json.dumps(data, default=dumps_default)
        post = dict(
            type=ttype,
            fire=fire
        )
        post.update(self.thread_data())
        return self.client.push('trace', post)

    def trace(self, action, data=dict()):
        data.update(type=action)
        return self.client.push('trace', data)

    def traceback_point(self, *args):
        tbs = extract_stack()
        tbs.reverse()
        fetch = False
        for file, line, func, t in tbs[1:]:
            if fetch:
                return dict(
                    file=file,
                    line=line,
                    message=args[0] if len(args)>0 else "no-message",
                    args=args
                )
            if func.lower() in ('log', 'warning', 'error'):
                fetch = True
                continue
        return None



from setuptools import setup

setup(
    name='xloger',
    version="1.3.0",
    url='https://github.com/idollo/xloger-python.git',
    license='MIT',
    author='idollo',
    author_email='stone58@qq.com',
    description='python client lib for xloger.',
    long_description=__doc__,
    packages=['xloger'],
    include_package_data=True,
    zip_safe=False,
    platforms='any',
    install_requires=[
        "socketpool>=0.5.3"
    ],
    classifiers=[
        'Environment :: Web Environment',
        'Intended Audience :: Developers',
        'Programming Language :: Python',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.3',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.5'
    ]
)
